#include "system.h"

#include <stdio.h> //Standard io
#include <stdlib.h> // VS_DSP Standard Library
#include <vs1000.h> // VS1000B register definitions
#include <vectors.h> // VS1000B vectors (interrupts and services)
#include <minifat.h> // Read Only Fat Filesystem
#include <mapper.h> // Logical Disk
#include <string.h> // memcpy etc
#include <player.h> // VS1000B default ROM player
#include <audio.h> // DAC output
#include <codec.h> // CODEC interface
#include <vsNand.h>
#include <mappertiny.h>
#include <usb.h>

#include <dev1000.h>

#include "gpioctrl.h"

extern struct CodecServices cs;

/*
  We currently allow two GPIO controls:
  1) GPIO_MASK for binary-encoded file number, all-zero means idle.
     The playing file is changed when the value changes to another non-zero
     value while a file is playing. The file is repeated if the number
     remains at the GPIO pins after the playback has finished.

     The number of bits used can be configured from 1 to 8 (GPIO0[0:7])
     for 255 files when SD card is used, and from 1 to 13 (GPIO0[0:12])
     for 8191 files when SD card is not used.

  2) GPIO_PRIORITIES for one GPIO per file with priorities.
     Lower number GPIO pin number has priority over a higher number GPIO pin.
     A higher number GPIO pin does not interrupt the playing of a current
     file, even if the lower number GPIO pin is no longer set.
     The file is repeated if a file is selected by the GPIO pins after the
     playback has finished.

     The number of GPIO pins used can be configured from 1 to 8 (GPIO0[0:7])
     when SD card is used, and from 1 to 13 (GPIO0[0:12]) when SD card is
     not used.

  Playing indicator:
     - 
  See system.h for GPIO_NUMBERED_FILES definition
 */

#if 0
#define GPIO_NODRIVE  // Do not drive pins to 0/1 before scanning -- requires external pull-downs

//#define GPIO_MASK       0x001f // Sel. file w/ GPIO0 pins in the mask
//#define GPIO0_PLAYING_INDICATOR  (1<<6)

//#define GPIO_MASK       0x0007 // Sel. file w/ GPIO0 pins in the mask
#define GPIO_INVERTED   //default state is HIGH, buttons bring low
#define GPIO_PRIORITIES   0x0007 // Sel. file w/ GPIO0 pins, lower has priority
#define NO_PRIORITIES //play to the end
//#define GPIO_PRIORITIES   0x04ff // Sel. file w/ GPIO0 pins, lower has priority
//#define STOP_WHEN_RELEASED /*stop playing when button released*/
#endif

/*
· Playing .WAV file number 0 with GPIO0 hold looping trigger i.e Instead of looking for rising edge to play, it should play ONLY as long as GPIO0 pulled up to IOVDD with a perfectly smooth transition between the end and beginning.
GPIO_MASK = 1
GPIO_NODRIVE
STOP_WHEN_RELEASED
GPIO_LOOPING

· Playing .WAV  file number 1 with GPIO1 trigger i.e if GPIO1 see rising edge player should play the whole file and stop, and in case GPIO sees rising edge while player is playing the file it should start the WAV file from beginning without any delay.
GPIO_RISING_EDGE = 2
GPIO_MASK
*/
#if 0
#define GPIO_NODRIVE
#define GPIO_MASK       0x0003
#define STOP_WHEN_RELEASED       /*GPIO0_0 - loop while '1' */
#define GPIO_RISING_EDGE 0x0002  /*GPIO0_1 - play once for rising edge */
#define GPIO_LOOPING
#define GPIO0_PLAYING_INDICATOR  0x0004 /*GPIO0_2 - '1' while playing */
#endif

#if 0 /* Innotech */
#define GPIO_INVERTED   /* active low */
#define GPIO_PRIORITIES  0x04ff /* GPIO0_0..7 and GPIO0_10 (CS1) */
#define GPIO_NODRIVE    /* require pull-ups, ignore floating. */
#endif

#ifdef LENCORE /*lencore*/
#define GPIO_NODRIVE
#define GPIO_INVERTED   //default state is HIGH, buttons bring low
#define GPIO_PRIORITIES 0x00ff
#define GPIO_LOOPING
#define GPIO0_PLAYING_INDICATOR  (1<<12)
//#define NO_PRIORITIES //play to the end
//#define STOP_WHEN_RELEASED /*stop playing when button released*/
#define GPIO_VOLUME_UP   (1<<13) /*vol+ GPIO0_13, vol- GPIO0_14, falling edge*/
#define GPIO_VOLUME_DOWN (1<<14) /*vol+ GPIO0_13, vol- GPIO0_14, falling edge*/
#define GPIO_VOLUME      ((GPIO_VOLUME_UP)|(GPIO_VOLUME_DOWN))
//autoprog/pcflash -p 2 -s 115200 -l spiall.spi -i autoprog/content.raw -r 64
#endif

#if 1
//#define GPIO_NODRIVE
#define GPIO_MASK       0x001f
#define STOP_WHEN_RELEASED       /*GPIO0_0 - loop while '1' */
//#define GPIO_RISING_EDGE 0x0002  /*GPIO0_1 - play once for rising edge */
//#define GPIO_LOOPING
//#define GPIO0_PLAYING_INDICATOR  0x0004 /*GPIO0_2 - '1' while playing */
#endif


#ifndef DEFAULT_FIRMWARE_KEYS
//#define GPIO0_PLAYING_INDICATOR  (1<<6)
#endif


#if defined(GPIO_MASK) && defined(GPIO_PRIORITIES)
err 'Both GPIO_MASK and GPIO_PRIORITIES can not be defined at the same time'
#endif



void GPIOCtrlIdleHook(void) {
  //USEX(UART_DATA) = '-';
#if defined(GPIO_VOLUME)
  {
    /* Do it separately from play request polling (Polling does bounce removal)
       Use GPIO hardware edge detection.
     */
    register s_int16 mask = PERIP(GPIO0_INT_PEND);
    PERIP(GPIO0_INT_PEND) = mask; /*Clear*/
    if ((mask & GPIO_VOLUME_UP)) {
      KeyEventHandler(ke_volumeUp2);
      //putword(player.volume);
    }
    if ((mask & GPIO_VOLUME_DOWN)) {
      KeyEventHandler(ke_volumeDown2);
      //putword(player.volume);
    }
  }
#endif /* GPIO_VOLUME */

#ifndef GPIO_NODRIVE
    if (uiTrigger) {
        uiTrigger = 0;
#endif
        /*
          If you want the default firmware key-controls, use KeyScan();
        */
#ifdef DEFAULT_FIRMWARE_KEYS
        KeyScan();
#else /*DEFAULT_FIRMWARE_KEYS*/


#if defined(GPIO0_PLAYING_INDICATOR)
        /* We do it this way so that the logic can be in this file. */
        if (player.currentFile != -1) {
            /* high = playing */
            PERIP(GPIO0_SET_MASK) = GPIO0_PLAYING_INDICATOR;
        } else {
            /* low = waiting for GPIO */
            PERIP(GPIO0_CLEAR_MASK) = GPIO0_PLAYING_INDICATOR;
        }
#endif

        /*TURN GPIO into GPIO mode.*/

/*--------------------------------------------------------------------*/
        
/*--------------------------------------------------------------------*/
#ifdef GPIO_MASK
        /* Direct selection through GPIO pins. */
        {
            register s_int16 mask;
#if defined(GPIO_RISING_EDGE)
            static s_int16 oldMask = -1;
#endif
#ifdef GPIO_INVERTED

#if !defined(GPIO_NODRIVE)
            /*default state is high, button pulls low to GND - except pull-ups*/
            PERIP(GPIO0_SET_MASK)   = GPIO_MASK | GPIO0_PULLUPS;
            PERIP(GPIO0_DDR) |= GPIO_MASK;  //Drive to 1
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
            PERIP(GPIO0_DDR) &= ~GPIO_MASK; //To inputs
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
#endif/*!GPIO_NODRIVE*/
            /* 20180213 In inverted mode everything is pull-ups. */
            mask = ((PERIP(GPIO0_IDATA) /*^ GPIO0_PULLUPS*/) & GPIO_MASK) ^ GPIO_MASK;
            
#else/*GPIO_INVERTED*/
#ifndef GPIO_NODRIVE
            /*default state is low, button pulls high to IOVDD*/
            PERIP(GPIO0_CLEAR_MASK) = GPIO_MASK;
            PERIP(GPIO0_SET_MASK)   = GPIO_MASK & GPIO0_PULLUPS;
            PERIP(GPIO0_DDR) |= GPIO_MASK;  //Drive to 0
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
            PERIP(GPIO0_DDR) &= ~GPIO_MASK; //To inputs
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
#endif/*!GPIO_NODRIVE*/
            mask = (PERIP(GPIO0_IDATA) ^ GPIO0_PULLUPS) & GPIO_MASK;
#endif/*elseGPIO_INVERTED*/
            //putch(mask);
            if (mask
#ifdef GPIO_RISING_EDGE
        	& ~GPIO_RISING_EDGE /* exclude these */
#endif
        	) {
        	/* If different file (or no file) is playing,
        	   cancel play and start playing the new one. */
        	if (
#ifdef GPIO_RISING_EDGE
        	    (player.currentFile == -1) /* does not override */
#else
        	    player.currentFile != mask - 1
#endif
        	    ) {
        	    player.currentFile = player.nextFile = mask - 1;
        	    cs.cancel = 1;
        	    player.pauseOn = 0;
#if defined(GPIO_LOOPING) && defined(LOOP_MODE)
        	    loopMode = 1;
#endif
        	}
#ifdef STOP_WHEN_RELEASED
            } else
#if defined(GPIO_RISING_EDGE) && defined(GPIO_LOOPING) && defined(LOOP_MODE)
              if (loopMode)
#endif
        	{
              player.pauseOn = 0;
              cs.cancel = 1;
#if defined(GPIO_LOOPING) && defined(LOOP_MODE)
              loopMode = 0;
              leaveLoop = 1;
#endif
#endif
            }
#ifdef GPIO_RISING_EDGE
            if (!(oldMask & GPIO_RISING_EDGE) &&
        	(mask & GPIO_RISING_EDGE)) {
              /* Rising edge - restart file as soon as possible. */
              player.pauseOn = 0;
              cs.cancel = 1;
              player.currentFile = player.nextFile = mask - 1;
#if defined(GPIO_LOOPING) && defined(LOOP_MODE)
              loopMode = 0;
#endif
            }
            oldMask = mask;
#endif
        }
#endif /*GPIO_MASK*/
/*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*/
#ifdef GPIO_PRIORITIES
        /* Prioritized selection through GPIO pins. */
        {
            register u_int16 mask;
#ifdef GPIO_INVERTED
            
#ifndef GPIO_NODRIVE
            PERIP(GPIO0_SET_MASK)   = GPIO_PRIORITIES /*& GPIO0_PULLUPS*/;
            PERIP(GPIO0_DDR) |= GPIO_PRIORITIES;  //Drive to 1
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
            PERIP(GPIO0_DDR) &= ~GPIO_PRIORITIES; //To inputs
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
#endif/*!GPIO_NODRIVE*/
            /* active low */
            mask = (PERIP(GPIO0_IDATA) & GPIO_PRIORITIES) ^ GPIO_PRIORITIES;
#else /*GPIO_INVERTED*/

#ifndef GPIO_NODRIVE
            PERIP(GPIO0_CLEAR_MASK) = GPIO_PRIORITIES;
            PERIP(GPIO0_SET_MASK)   = GPIO_PRIORITIES & GPIO0_PULLUPS;
            PERIP(GPIO0_DDR) |= GPIO_PRIORITIES;  //Drive to 0
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
            PERIP(GPIO0_DDR) &= ~GPIO_PRIORITIES; //To inputs
            for (mask = 0;mask<10000;mask++)
        	USEX(0);
#endif/*!GPIO_NODRIVE*/
            /*Active high*/
            mask = (PERIP(GPIO0_IDATA) ^ GPIO0_PULLUPS) & GPIO_PRIORITIES;
#endif/*else GPIO_INVERTED*/

#ifdef LENCORE /*lencore*/
            //if (mask)	      putword(mask);
#endif
            if (mask) {
        	register s_int16 n = 0, m = GPIO_PRIORITIES;
        	//putch(mask);
        	while (m) {
        	    if (m & 1) {
        		if (mask & 1) {
        		    break;
        		}
        		n++;
        	    }
        	    m >>= 1;
        	    mask >>= 1;
        	}
        	if (
#ifndef NO_PRIORITIES /* one pin per file, lower GPIO numbers have priority */
        	    player.currentFile == -1 || n < player.currentFile
#else /* one pin per file */
        	    n != player.currentFile
#endif
        	    ) {
        	    player.currentFile = player.nextFile = n;
        	    cs.cancel = 1;
        	    player.pauseOn = 0;
        	}
#ifdef STOP_WHEN_RELEASED
            } else if (player.currentFile != -1) {
              player.pauseOn = 0;
              cs.cancel = 1;
#endif
            }
        }
#endif /*GPIO_PRIORITIES*/
/*--------------------------------------------------------------------*/


#if !defined(GPIO_MASK) && !defined(GPIO_PRIORITIES)
        /* Buttons */
        //KeyScanNoUSB(0x1f); /* with feature button */
#endif

#ifdef GPIO_NODRIVE
    if (uiTrigger) {
        uiTrigger = 0;
#endif
/*--------------------------------------------------------------------*/
#ifdef USE_POWERBUTTON
        if (PERIP(SCI_STATUS) & SCISTF_REGU_POWERBUT) {
            /* In 2FIN2 switching on the uSD regulator affects powerbutton. */
            if (keyOld == (SCISTF_REGU_POWERBUT<<4)) {
        	if (++keyOldTime >= 3) { /* about 0.2 sec */
        	    PERIP(GPIO0_CLEAR_MASK) = GPIO0_SD_POWER;
        	    RealPowerOff();
        	} else {
        	    keyOldTime = 0;
        	}
            } else {
        	keyOld == (SCISTF_REGU_POWERBUT<<4);
            }
        } else {
            keyOld = 0;
        }
#endif
/*--------------------------------------------------------------------*/
#endif /*DEFAULT_FIRMWARE_KEYS*/


#ifndef DISABLE_USB
        if (USBIsAttached()) {
            cs.cancel = 1;
            player.pauseOn = 0;
        }
#endif
        CheckSd();
    }
}

void GPIOInit(void) {
  /* GPIO_MASK and GPIO_PRIORITIES will override playing indicator
     if they use the same pins. */
#ifdef GPIO0_PLAYING_INDICATOR
  PERIP(GPIO0_MODE) &= ~GPIO0_PLAYING_INDICATOR;
  PERIP(GPIO0_DDR) |= GPIO0_PLAYING_INDICATOR;
#endif
  /*TURN GPIO into GPIO mode.*/
#ifdef GPIO_MASK
  PERIP(GPIO0_MODE) &= ~GPIO_MASK;
  PERIP(GPIO0_DDR)  &= ~GPIO_MASK; /* GPIO0[0:n] into inputs */
#endif /*GPIO_MASK*/
#ifdef GPIO_PRIORITIES
  PERIP(GPIO0_MODE) &= ~GPIO_PRIORITIES;
  PERIP(GPIO0_DDR)  &= ~GPIO_PRIORITIES; /* GPIO0[0:n] into inputs */
#endif /*GPIO_PRIORITIES*/
#ifdef GPIO_VOLUME
  PERIP(GPIO0_MODE) &= ~GPIO_VOLUME;
  PERIP(GPIO0_DDR)  &= ~GPIO_VOLUME;
  PERIP(GPIO0_INT_FALL) |= GPIO_VOLUME;
  PERIP(GPIO0_INT_PEND)  = GPIO_VOLUME; /*Clear*/
#endif
  player.nextFile = -1; //No file is being played when we start
  player.volume = -24; //Max volume
}
